@@ -2,7 +2,7 @@
 
 /*  vcfplugin.c -- plugin modules for operating on VCF/BCF files.
 
-    Copyright (C) 2013-2017 Genome Research Ltd.
+    Copyright (C) 2013-2021 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -139,7 +139,7 @@ typedef struct _args_t
     bcf_srs_t *files;
     bcf_hdr_t *hdr, *hdr_out;
     htsFile *out_fh;
-    int output_type, n_threads;
+    int output_type, n_threads, clevel;
 
     filter_t *filter;
     char *filter_str;
@@ -150,7 +150,7 @@ typedef struct _args_t
     char **plugin_paths;
 
     char **argv, *output_fname, *regions_list, *targets_list;
-    int argc, drop_header, verbose, record_cmd_line;
+    int argc, drop_header, verbose, record_cmd_line, plist_only;
 }
 args_t;
 
@@ -180,7 +180,7 @@ static void add_plugin_paths(args_t *args, const char *path)
                 args->plugin_paths = (char**) realloc(args->plugin_paths,sizeof(char*)*(args->nplugin_paths+1));
                 args->plugin_paths[args->nplugin_paths] = dir;
                 args->nplugin_paths++;
-                if ( args->verbose > 1 ) fprintf(bcftools_stderr, "plugin directory %s .. ok\n", dir);
+                if ( args->verbose > 1 && strcmp(".",dir) ) fprintf(bcftools_stderr, "plugin directory %s .. ok\n", dir);
             }
             else
             {
@@ -222,6 +222,8 @@ static void *dlopen_plugin(args_t *args, const char *fname)
 #else
     if ( fname[0]=='/' ) is_absolute_path = 1;
 #endif
+
+    kstring_t err = {0,0,0};
     if ( !is_absolute_path )
     {
         int i;
@@ -233,16 +235,14 @@ static void *dlopen_plugin(args_t *args, const char *fname)
 #else
             handle = dlopen(tmp, RTLD_NOW); // valgrind complains about unfreed memory, not our problem though
 #endif
-            if ( args->verbose > 1 )
-            {
-                if ( !handle )
+            if ( !handle )
 #ifdef _WIN32
-                    fprintf(bcftools_stderr,"%s:\n\tLoadLibraryA   .. %lu\n", tmp, GetLastError());
+                ksprintf(&err,"LoadLibraryA   .. %lu\n", GetLastError());
 #else
-                    fprintf(bcftools_stderr,"%s:\n\tdlopen   .. %s\n", tmp, dlerror());
+                ksprintf(&err,"%s:\n\tdlopen   .. %s\n", tmp,dlerror());
 #endif
-                else fprintf(bcftools_stderr,"%s:\n\tplugin open   .. ok\n", tmp);
-            }
+            else if ( args->verbose > 1 )
+                fprintf(bcftools_stderr,"%s:\n\tplugin open   .. ok\n", tmp);
             free(tmp);
             if ( handle ) return handle;
         }
@@ -253,33 +253,46 @@ static void *dlopen_plugin(args_t *args, const char *fname)
 #else
     handle = dlopen(fname, RTLD_NOW);
 #endif
-    if ( args->verbose > 1 )
-    {
-        if ( !handle )
+    if ( !handle )
 #ifdef _WIN32
-            fprintf(bcftools_stderr,"%s:\n\tLoadLibraryA   .. %lu\n", fname, GetLastError());
+        ksprintf(&err,"LoadLibraryA   .. %lu\n", GetLastError());
 #else
-            fprintf(bcftools_stderr,"%s:\n\tdlopen   .. %s\n", fname, dlerror());
+        ksprintf(&err,"%s:\n\tdlopen   .. %s\n", fname,dlerror());
 #endif
-        else fprintf(bcftools_stderr,"%s:\n\tplugin open   .. ok\n", fname);
-    }
+    else if ( args->verbose > 1 )
+        fprintf(bcftools_stderr,"%s:\n\tplugin open   .. ok\n", fname);
+
+    if ( !handle && (!args->plist_only || args->verbose>1) )
+        fprintf(bcftools_stderr,"%s",err.s);
+    free(err.s);
 
     return handle;
 }
 
-static void print_plugin_usage_hint(void)
+static void print_plugin_usage_hint(const char *name)
 {
-    fprintf(bcftools_stderr, "\nNo functional bcftools plugins were found");
-    if ( !getenv("BCFTOOLS_PLUGINS") )
-        fprintf(bcftools_stderr,". The environment variable BCFTOOLS_PLUGINS is not set.\n\n");
+    if ( name )
+        fprintf(bcftools_stderr, "\nThe bcftools plugin \"%s\" was not found or is not functional", name);
     else
+        fprintf(bcftools_stderr, "\nNo functional bcftools plugins were found");
+    if ( !getenv("BCFTOOLS_PLUGINS") )
+    {
+        fprintf(bcftools_stderr,". The environment variable BCFTOOLS_PLUGINS is not set");
+#ifdef PLUGINPATH
+        fprintf(bcftools_stderr,"\nand no usable plugins were found in %s", PLUGINPATH);
+#endif
+        fprintf(bcftools_stderr,".\n\n");
+    }
+    else
+    {
         fprintf(bcftools_stderr,
                 " in\n\tBCFTOOLS_PLUGINS=\"%s\".\n\n"
                 "- Is the plugin path correct?\n\n"
-                "- Run \"bcftools plugin -lv\" for more detailed error output.\n"
+                "- Run \"bcftools plugin -l\" or \"bcftools plugin -lvv\" for a list of available plugins.\n"
                 "\n",
                 getenv("BCFTOOLS_PLUGINS")
                );
+    }
 }
 
 static int load_plugin(args_t *args, const char *fname, int exit_on_error, plugin_t *plugin)
@@ -291,7 +304,7 @@ static int load_plugin(args_t *args, const char *fname, int exit_on_error, plugi
     {
         if ( exit_on_error )
         {
-            print_plugin_usage_hint();
+            print_plugin_usage_hint(fname);
             error("Could not load \"%s\".\n\n", fname);
         }
         return -1;
@@ -412,12 +425,9 @@ static int load_plugin(args_t *args, const char *fname, int exit_on_error, plugi
     return 0;
 }
 
-static void init_plugin(args_t *args)
+static void check_version(args_t *args)
 {
     static int warned_bcftools = 0, warned_htslib = 0;
-
-    int ret = args->plugin.init(args->plugin.argc,args->plugin.argv,args->hdr,args->hdr_out);
-    if ( ret<0 ) error("The plugin exited with an error.\n");
     const char *bver, *hver;
     args->plugin.version(&bver, &hver);
     if ( strcmp(bver,bcftools_version()) && !warned_bcftools )
@@ -430,6 +440,13 @@ static void init_plugin(args_t *args)
         fprintf(bcftools_stderr,"WARNING: htslib version mismatch .. bcftools at %s, the plugin \"%s\" at %s\n", hts_version(),args->plugin.name,hver);
         warned_htslib = 1;
     }
+}
+
+static void init_plugin(args_t *args)
+{
+    int ret = args->plugin.init(args->plugin.argc,args->plugin.argv,args->hdr,args->hdr_out);
+    if ( ret<0 ) error("The plugin exited with an error.\n");
+    check_version(args);
     args->drop_header += ret;
 }
 
@@ -489,7 +506,7 @@ static int list_plugins(args_t *args)
         if ( args->verbose ) fprintf(bcftools_stdout, "\n");
     }
     else
-        print_plugin_usage_hint();
+        print_plugin_usage_hint(NULL);
     free(str.s);
     return nplugins ? 0 : 1;
 }
@@ -507,7 +524,9 @@ static void init_data(args_t *args)
     if (args->record_cmd_line) bcf_hdr_append_version(args->hdr_out, args->argc, args->argv, "bcftools_plugin");
     if ( !args->drop_header )
     {
-        args->out_fh = hts_open(args->output_fname,hts_bcf_wmode(args->output_type));
+        char wmode[8];
+        set_wmode(wmode,args->output_type,args->output_fname,args->clevel);
+        args->out_fh = hts_open(args->output_fname ? args->output_fname : "-", wmode);
         if ( args->out_fh == NULL ) error("Can't write to \"%s\": %s\n", args->output_fname, strerror(errno));
         if ( args->n_threads ) hts_set_threads(args->out_fh, args->n_threads);
         if ( bcf_hdr_write(args->out_fh, args->hdr_out)!=0 ) error("[%s] Error: cannot write to %s\n", __func__,args->output_fname);
@@ -543,24 +562,26 @@ static void usage(args_t *args)
     fprintf(bcftools_stderr, "         bcftools +name [OPTIONS] <file>  [-- PLUGIN_OPTIONS]\n");
     fprintf(bcftools_stderr, "\n");
     fprintf(bcftools_stderr, "VCF input options:\n");
-    fprintf(bcftools_stderr, "   -e, --exclude <expr>        exclude sites for which the expression is true\n");
-    fprintf(bcftools_stderr, "   -i, --include <expr>        select sites for which the expression is true\n");
-    fprintf(bcftools_stderr, "   -r, --regions <region>      restrict to comma-separated list of regions\n");
-    fprintf(bcftools_stderr, "   -R, --regions-file <file>   restrict to regions listed in a file\n");
-    fprintf(bcftools_stderr, "   -t, --targets <region>      similar to -r but streams rather than index-jumps\n");
-    fprintf(bcftools_stderr, "   -T, --targets-file <file>   similar to -R but streams rather than index-jumps\n");
+    fprintf(bcftools_stderr, "   -e, --exclude EXPR             Exclude sites for which the expression is true\n");
+    fprintf(bcftools_stderr, "   -i, --include EXPR             Select sites for which the expression is true\n");
+    fprintf(bcftools_stderr, "   -r, --regions REGION           Restrict to comma-separated list of regions\n");
+    fprintf(bcftools_stderr, "   -R, --regions-file FILE        Restrict to regions listed in a file\n");
+    fprintf(bcftools_stderr, "        --regions-overlap 0|1|2   Include if POS in the region (0), record overlaps (1), variant overlaps (2) [1]\n");
+    fprintf(bcftools_stderr, "   -t, --targets REGION           Similar to -r but streams rather than index-jumps\n");
+    fprintf(bcftools_stderr, "   -T, --targets-file FILE        Similar to -R but streams rather than index-jumps\n");
+    fprintf(bcftools_stderr, "        --targets-overlap 0|1|2   Include if POS in the region (0), record overlaps (1), variant overlaps (2) [0]\n");
     fprintf(bcftools_stderr, "VCF output options:\n");
-    fprintf(bcftools_stderr, "       --no-version            do not append version and command line to the header\n");
-    fprintf(bcftools_stderr, "   -o, --output <file>         write output to a file [standard output]\n");
-    fprintf(bcftools_stderr, "   -O, --output-type <type>    'b' compressed BCF; 'u' uncompressed BCF; 'z' compressed VCF; 'v' uncompressed VCF [v]\n");
-    fprintf(bcftools_stderr, "       --threads <int>         use multithreading with <int> worker threads [0]\n");
+    fprintf(bcftools_stderr, "       --no-version               Do not append version and command line to the header\n");
+    fprintf(bcftools_stderr, "   -o, --output FILE              Write output to a file [standard output]\n");
+    fprintf(bcftools_stderr, "   -O, --output-type u|b|v|z[0-9] u/b: un/compressed BCF, v/z: un/compressed VCF, 0-9: compression level [v]\n");
+    fprintf(bcftools_stderr, "       --threads INTT             Use multithreading with <int> worker threads [0]\n");
     fprintf(bcftools_stderr, "Plugin options:\n");
-    fprintf(bcftools_stderr, "   -h, --help                  list plugin's options\n");
-    fprintf(bcftools_stderr, "   -l, --list-plugins          list available plugins. See BCFTOOLS_PLUGINS environment variable and man page for details\n");
-    fprintf(bcftools_stderr, "   -v, --verbose               print verbose information, -vv increases verbosity\n");
-    fprintf(bcftools_stderr, "   -V, --version               print version string and exit\n");
+    fprintf(bcftools_stderr, "   -h, --help                     List plugin's options\n");
+    fprintf(bcftools_stderr, "   -l, --list-plugins             List available plugins. See BCFTOOLS_PLUGINS environment variable and man page for details\n");
+    fprintf(bcftools_stderr, "   -v, --verbose                  Print verbose information, -vv increases verbosity\n");
+    fprintf(bcftools_stderr, "   -V, --version                  Print version string and exit\n");
     fprintf(bcftools_stderr, "\n");
-    exit(1);
+    bcftools_exit(1);
 }
 
 static int is_verbose(int argc, char *argv[])
@@ -594,10 +615,12 @@ int main_plugin(int argc, char *argv[])
     args->n_threads = 0;
     args->record_cmd_line = 1;
     args->nplugin_paths = -1;
-    int regions_is_file = 0, targets_is_file = 0, plist_only = 0, usage_only = 0, version_only = 0;
+    args->clevel = -1;
+    int regions_is_file = 0, targets_is_file = 0, usage_only = 0, version_only = 0;
+    int regions_overlap = 1;
+    int targets_overlap = 0;
 
     if ( argc==1 ) usage(args);
-
     char *plugin_name = NULL;
     if ( argv[1][0]!='-' )
     {
@@ -608,6 +631,7 @@ int main_plugin(int argc, char *argv[])
         load_plugin(args, plugin_name, 1, &args->plugin);
         if ( args->plugin.run )
         {
+            check_version(args);
             int ret = args->plugin.run(argc, argv);
             destroy_data(args);
             free(args);
@@ -628,11 +652,14 @@ int main_plugin(int argc, char *argv[])
         {"exclude",required_argument,NULL,'e'},
         {"regions",required_argument,NULL,'r'},
         {"regions-file",required_argument,NULL,'R'},
+        {"regions-overlap",required_argument,NULL,1},
         {"targets",required_argument,NULL,'t'},
         {"targets-file",required_argument,NULL,'T'},
+        {"targets-overlap",required_argument,NULL,2},
         {"no-version",no_argument,NULL,8},
         {NULL,0,NULL,0}
     };
+    char *tmp;
     while ((c = getopt_long(argc, argv, "h?o:O:r:R:t:T:li:e:vV",loptions,NULL)) >= 0)
     {
         switch (c) {
@@ -645,16 +672,41 @@ int main_plugin(int argc, char *argv[])
                     case 'u': args->output_type = FT_BCF; break;
                     case 'z': args->output_type = FT_VCF_GZ; break;
                     case 'v': args->output_type = FT_VCF; break;
-                    default: error("The output type \"%s\" not recognised\n", optarg);
+                    default:
+                    {
+                        args->clevel = strtol(optarg,&tmp,10);
+                        if ( *tmp || args->clevel<0 || args->clevel>9 ) error("The output type \"%s\" not recognised\n", optarg);
+                    }
                 };
+                if ( optarg[1] )
+                {
+                    args->clevel = strtol(optarg+1,&tmp,10);
+                    if ( *tmp || args->clevel<0 || args->clevel>9 ) error("Could not parse argument: --compression-level %s\n", optarg+1);
+                }
                 break;
-            case 'e': args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
-            case 'i': args->filter_str = optarg; args->filter_logic |= FLT_INCLUDE; break;
+            case 'e':
+                if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
+                args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
+            case 'i':
+                if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
+                args->filter_str = optarg; args->filter_logic |= FLT_INCLUDE; break;
             case 'r': args->regions_list = optarg; break;
             case 'R': args->regions_list = optarg; regions_is_file = 1; break;
             case 't': args->targets_list = optarg; break;
             case 'T': args->targets_list = optarg; targets_is_file = 1; break;
-            case 'l': plist_only = 1; break;
+            case 'l': args->plist_only = 1; break;
+            case  1 :
+                if ( !strcasecmp(optarg,"0") ) regions_overlap = 0;
+                else if ( !strcasecmp(optarg,"1") ) regions_overlap = 1;
+                else if ( !strcasecmp(optarg,"2") ) regions_overlap = 2;
+                else error("Could not parse: --regions-overlap %s\n",optarg);
+                break;
+            case  2 :
+                if ( !strcasecmp(optarg,"0") ) targets_overlap = 0;
+                else if ( !strcasecmp(optarg,"1") ) targets_overlap = 1;
+                else if ( !strcasecmp(optarg,"2") ) targets_overlap = 2;
+                else error("Could not parse: --targets-overlap %s\n",optarg);
+                break;
             case  9 : args->n_threads = strtol(optarg, 0, 0); break;
             case  8 : args->record_cmd_line = 0; break;
             case '?':
@@ -662,8 +714,8 @@ int main_plugin(int argc, char *argv[])
             default: error("Unknown argument: %s\n", optarg);
         }
     }
-    if ( plist_only )  return list_plugins(args);
-    if ( usage_only && ! plugin_name ) usage(args);
+    if ( args->plist_only )  return list_plugins(args);
+    if ( !plugin_name ) usage(args);
 
     if ( version_only )
     {
@@ -684,7 +736,7 @@ int main_plugin(int argc, char *argv[])
     }
 
     char *fname = NULL;
-    if ( optind>=argc || argv[optind][0]=='-' )
+    if ( optind>=argc || (argv[optind][0]=='-' && argv[optind][1]) )
     {
         args->plugin.argc = argc - optind + 1;
         args->plugin.argv = argv + optind - 1;
@@ -708,11 +760,13 @@ int main_plugin(int argc, char *argv[])
     args->files = bcf_sr_init();
     if ( args->regions_list )
     {
+        bcf_sr_set_opt(args->files,BCF_SR_REGIONS_OVERLAP,regions_overlap);
         if ( bcf_sr_set_regions(args->files, args->regions_list, regions_is_file)<0 )
             error("Failed to read the regions: %s\n", args->regions_list);
     }
     if ( args->targets_list )
     {
+        bcf_sr_set_opt(args->files,BCF_SR_TARGETS_OVERLAP,targets_overlap);
         if ( bcf_sr_set_targets(args->files, args->targets_list, targets_is_file, 0)<0 )
             error("Failed to read the targets: %s\n", args->targets_list);
         args->files->collapse |= COLLAPSE_SOME;
