@@ -2,7 +2,7 @@
 
 /*  vcfquery.c -- Extracts fields from VCF/BCF file.
 
-    Copyright (C) 2013-2017 Genome Research Ltd.
+    Copyright (C) 2013-2021 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -223,25 +223,27 @@ static void usage(void)
     fprintf(bcftools_stderr, "Usage:   bcftools query [options] <A.vcf.gz> [<B.vcf.gz> [...]]\n");
     fprintf(bcftools_stderr, "\n");
     fprintf(bcftools_stderr, "Options:\n");
-    fprintf(bcftools_stderr, "    -e, --exclude <expr>              exclude sites for which the expression is true (see man page for details)\n");
-    fprintf(bcftools_stderr, "    -f, --format <string>             see man page for details\n");
-    fprintf(bcftools_stderr, "    -H, --print-header                print header\n");
-    fprintf(bcftools_stderr, "    -i, --include <expr>              select sites for which the expression is true (see man page for details)\n");
-    fprintf(bcftools_stderr, "    -l, --list-samples                print the list of samples and exit\n");
-    fprintf(bcftools_stderr, "    -o, --output-file <file>          output file name [bcftools_stdout]\n");
-    fprintf(bcftools_stderr, "    -r, --regions <region>            restrict to comma-separated list of regions\n");
-    fprintf(bcftools_stderr, "    -R, --regions-file <file>         restrict to regions listed in a file\n");
-    fprintf(bcftools_stderr, "    -s, --samples <list>              list of samples to include\n");
-    fprintf(bcftools_stderr, "    -S, --samples-file <file>         file of samples to include\n");
-    fprintf(bcftools_stderr, "    -t, --targets <region>            similar to -r but streams rather than index-jumps\n");
-    fprintf(bcftools_stderr, "    -T, --targets-file <file>         similar to -R but streams rather than index-jumps\n");
-    fprintf(bcftools_stderr, "    -u, --allow-undef-tags            print \".\" for undefined tags\n");
-    fprintf(bcftools_stderr, "    -v, --vcf-list <file>             process multiple VCFs listed in the file\n");
+    fprintf(bcftools_stderr, "    -e, --exclude EXPR                Exclude sites for which the expression is true (see man page for details)\n");
+    fprintf(bcftools_stderr, "    -f, --format STRING               See man page for details\n");
+    fprintf(bcftools_stderr, "    -H, --print-header                Print header\n");
+    fprintf(bcftools_stderr, "    -i, --include EXPR                Select sites for which the expression is true (see man page for details)\n");
+    fprintf(bcftools_stderr, "    -l, --list-samples                Print the list of samples and exit\n");
+    fprintf(bcftools_stderr, "    -o, --output FILE                 Output file name [bcftools_stdout]\n");
+    fprintf(bcftools_stderr, "    -r, --regions REGION              Restrict to comma-separated list of regions\n");
+    fprintf(bcftools_stderr, "    -R, --regions-file FILE           Restrict to regions listed in a file\n");
+    fprintf(bcftools_stderr, "        --regions-overlap 0|1|2       Include if POS in the region (0), record overlaps (1), variant overlaps (2) [1]\n");
+    fprintf(bcftools_stderr, "    -s, --samples LIST                List of samples to include\n");
+    fprintf(bcftools_stderr, "    -S, --samples-file FILE           File of samples to include\n");
+    fprintf(bcftools_stderr, "    -t, --targets REGION              Similar to -r but streams rather than index-jumps\n");
+    fprintf(bcftools_stderr, "    -T, --targets-file FILE           Similar to -R but streams rather than index-jumps\n");
+    fprintf(bcftools_stderr, "        --targets-overlap 0|1|2       Include if POS in the region (0), record overlaps (1), variant overlaps (2) [0]\n");
+    fprintf(bcftools_stderr, "    -u, --allow-undef-tags            Print \".\" for undefined tags\n");
+    fprintf(bcftools_stderr, "    -v, --vcf-list FILE               Process multiple VCFs listed in the file\n");
     fprintf(bcftools_stderr, "\n");
     fprintf(bcftools_stderr, "Examples:\n");
     fprintf(bcftools_stderr, "\tbcftools query -f '%%CHROM\\t%%POS\\t%%REF\\t%%ALT[\\t%%SAMPLE=%%GT]\\n' file.vcf.gz\n");
     fprintf(bcftools_stderr, "\n");
-    exit(1);
+    bcftools_exit(1);
 }
 
 int main_vcfquery(int argc, char *argv[])
@@ -250,6 +252,8 @@ int main_vcfquery(int argc, char *argv[])
     args_t *args = (args_t*) calloc(1,sizeof(args_t));
     args->argc   = argc; args->argv = argv;
     int regions_is_file = 0, targets_is_file = 0;
+    int regions_overlap = 1;
+    int targets_overlap = 0;
 
     static struct option loptions[] =
     {
@@ -259,10 +263,13 @@ int main_vcfquery(int argc, char *argv[])
         {"exclude",1,0,'e'},
         {"format",1,0,'f'},
         {"output-file",1,0,'o'},
+        {"output",1,0,'o'},
         {"regions",1,0,'r'},
         {"regions-file",1,0,'R'},
+        {"regions-overlap",required_argument,NULL,1},
         {"targets",1,0,'t'},
         {"targets-file",1,0,'T'},
+        {"targets-overlap",required_argument,NULL,2},
         {"annots",1,0,'a'},
         {"samples",1,0,'s'},
         {"samples-file",1,0,'S'},
@@ -298,8 +305,12 @@ int main_vcfquery(int argc, char *argv[])
                     args->format_str = str.s;
                     break;
                 }
-            case 'e': args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
-            case 'i': args->filter_str = optarg; args->filter_logic |= FLT_INCLUDE; break;
+            case 'e':
+                if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
+                args->filter_str = optarg; args->filter_logic |= FLT_EXCLUDE; break;
+            case 'i':
+                if ( args->filter_str ) error("Error: only one -i or -e expression can be given, and they cannot be combined\n");
+                args->filter_str = optarg; args->filter_logic |= FLT_INCLUDE; break;
             case 'r': args->regions_list = optarg; break;
             case 'R': args->regions_list = optarg; regions_is_file = 1; break;
             case 't': args->targets_list = optarg; break;
@@ -308,6 +319,18 @@ int main_vcfquery(int argc, char *argv[])
             case 'u': args->allow_undef_tags = 1; break;
             case 's': args->sample_list = optarg; break;
             case 'S': args->sample_list = optarg; args->sample_is_file = 1; break;
+            case  1 :
+                if ( !strcasecmp(optarg,"0") ) regions_overlap = 0;
+                else if ( !strcasecmp(optarg,"1") ) regions_overlap = 1;
+                else if ( !strcasecmp(optarg,"2") ) regions_overlap = 2;
+                else error("Could not parse: --regions-overlap %s\n",optarg);
+                break;
+            case  2 :
+                if ( !strcasecmp(optarg,"0") ) targets_overlap = 0;
+                else if ( !strcasecmp(optarg,"1") ) targets_overlap = 1;
+                else if ( !strcasecmp(optarg,"2") ) targets_overlap = 2;
+                else error("Could not parse: --targets-overlap %s\n",optarg);
+                break;
             case 'h':
             case '?': usage(); break;
             default: error("Unknown argument: %s\n", optarg);
@@ -345,10 +368,15 @@ int main_vcfquery(int argc, char *argv[])
         if ( !fname ) usage();
         args->files = bcf_sr_init();
         if ( optind+1 < argc ) args->files->require_index = 1;
-        if ( args->regions_list && bcf_sr_set_regions(args->files, args->regions_list, regions_is_file)<0 )
-            error("Failed to read the regions: %s\n", args->regions_list);
+        if ( args->regions_list )
+        {
+            bcf_sr_set_opt(args->files,BCF_SR_REGIONS_OVERLAP,regions_overlap);
+            if ( bcf_sr_set_regions(args->files, args->regions_list, regions_is_file)<0 )
+                error("Failed to read the regions: %s\n", args->regions_list);
+        }
         if ( args->targets_list )
         {
+            bcf_sr_set_opt(args->files,BCF_SR_TARGETS_OVERLAP,targets_overlap);
             if ( bcf_sr_set_targets(args->files, args->targets_list, targets_is_file, 0)<0 )
                 error("Failed to read the targets: %s\n", args->targets_list);
         }
