@@ -37,6 +37,11 @@ SAMTOOLS_DISPATCH = {
     "quickcheck": ("quickcheck", None),
     "split": ("split", None),
     "flags": ("flags", None),
+    "ampliconclip": ("ampliconclip", None),
+    "ampliconstats": ("ampliconstats", None),
+    "version": ("version", None),
+    "fqimport": ("import", None),
+    "samples": ("samples", None),
 }
 
 # instantiate samtools commands as python functions
