@@ -2,7 +2,7 @@
 
 /*  bam_plcmd.c -- mpileup subcommand.
 
-    Copyright (C) 2008-2015, 2019 Genome Research Ltd.
+    Copyright (C) 2008-2015, 2019-2021 Genome Research Ltd.
     Portions copyright (C) 2009-2012 Broad Institute.
 
     Author: Heng Li <lh3@sanger.ac.uk>
@@ -44,6 +44,7 @@ DEALINGS IN THE SOFTWARE.  */
 #include <htslib/kstring.h>
 #include <htslib/klist.h>
 #include <htslib/khash_str2int.h>
+#include <htslib/cram.h>
 #include "samtools.h"
 #include "bedidx.h"
 #include "sam_opts.h"
@@ -67,11 +68,14 @@ static inline int printw(int c, FILE *fp)
 }
 
 static inline int pileup_seq(FILE *fp, const bam_pileup1_t *p, hts_pos_t pos,
-                              hts_pos_t ref_len, const char *ref, kstring_t *ks,
-                              int rev_del)
+                             hts_pos_t ref_len, const char *ref, kstring_t *ks,
+                             int rev_del, int no_ins, int no_ins_mods,
+                             int no_del, int no_ends)
 {
+    no_ins_mods |= no_ins;
     int j;
-    if (p->is_head) {
+    hts_base_mod_state *m = p->cd.p;
+    if (!no_ends && p->is_head) {
         putc('^', fp);
         putc(p->b->core.qual > 93? 126 : p->b->core.qual + 33, fp);
     }
@@ -88,30 +92,74 @@ static inline int pileup_seq(FILE *fp, const bam_pileup1_t *p, hts_pos_t pos,
             else c = bam_is_rev(p->b)? tolower(c) : toupper(c);
         }
         putc(c, fp);
+        if (m) {
+            int nm;
+            hts_base_mod mod[256];
+            if ((nm = bam_mods_at_qpos(p->b, p->qpos, m, mod, 256)) > 0) {
+                putc('[', fp);
+                int j;
+                for (j = 0; j < nm && j < 256; j++) {
+                    char qual[20];
+                    if (mod[j].qual >= 0)
+                        sprintf(qual, "%d", mod[j].qual);
+                    else
+                        *qual = 0;
+                    if (mod[j].modified_base < 0)
+                        // ChEBI
+                        fprintf(fp, "%c(%d)%s", "+-"[mod[j].strand],
+                                -mod[j].modified_base, qual);
+                    else
+                        fprintf(fp, "%c%c%s", "+-"[mod[j].strand],
+                                mod[j].modified_base, qual);
+                }
+                putc(']', fp);
+            }
+        }
     } else putc(p->is_refskip? (bam_is_rev(p->b)? '<' : '>') : ((bam_is_rev(p->b) && rev_del) ? '#' : '*'), fp);
     int del_len = -p->indel;
     if (p->indel > 0) {
-        int len = bam_plp_insertion(p, ks, &del_len);
-        if (len < 0)
+        int len = bam_plp_insertion_mod(p, m && !no_ins_mods ? m : NULL,
+                                        ks, &del_len);
+        if (len < 0) {
+            print_error("mpileup", "bam_plp_insertion() failed");
             return -1;
-        putc('+', fp); printw(len, fp);
-        if (bam_is_rev(p->b)) {
-            char pad = rev_del ? '#' : '*';
-            for (j = 0; j < len; j++)
-                putc(ks->s[j] != '*' ? tolower(ks->s[j]) : pad, fp);
-        } else {
-            for (j = 0; j < len; j++)
-                putc(toupper(ks->s[j]), fp);
+        }
+        if (no_ins < 2) {
+            putc('+', fp);
+            printw(len, fp);
+        }
+        if (!no_ins) {
+            if (bam_is_rev(p->b)) {
+                char pad = rev_del ? '#' : '*';
+                int in_mod = 0;
+                for (j = 0; j < ks->l; j++) {
+                    if (ks->s[j] == '[') in_mod = 1;
+                    else if (ks->s[j] == ']') in_mod = 0;
+                    putc(ks->s[j] != '*'
+                         ? (in_mod ? ks->s[j] : tolower(ks->s[j]))
+                         : pad, fp);
+                }
+            } else {
+                int in_mod = 0;
+                for (j = 0; j < ks->l; j++) {
+                    if (ks->s[j] == '[') in_mod = 1;
+                    if (ks->s[j] == ']') in_mod = 0;
+                    putc(in_mod ? ks->s[j] : toupper(ks->s[j]), fp);
+                }
+            }
         }
     }
     if (del_len > 0) {
-        printw(-del_len, fp);
-        for (j = 1; j <= del_len; ++j) {
-            int c = (ref && (int)pos+j < ref_len)? ref[pos+j] : 'N';
-            putc(bam_is_rev(p->b)? tolower(c) : toupper(c), fp);
+        if (no_del < 2)
+            printw(-del_len, fp);
+        if (!no_del) {
+            for (j = 1; j <= del_len; ++j) {
+                int c = (ref && (int)pos+j < ref_len)? ref[pos+j] : 'N';
+                putc(bam_is_rev(p->b)? tolower(c) : toupper(c), fp);
+            }
         }
     }
-    if (p->is_tail) putc('$', fp);
+    if (!no_ends && p->is_tail) putc('$', fp);
     return 0;
 }
 
@@ -128,10 +176,11 @@ static inline int pileup_seq(FILE *fp, const bam_pileup1_t *p, hts_pos_t pos,
 #define MPLP_REDO_BAQ   (1<<6)
 #define MPLP_ILLUMINA13 (1<<7)
 #define MPLP_IGNORE_RG  (1<<8)
-#define MPLP_PRINT_QPOS (1<<9)
-#define MPLP_PER_SAMPLE (1<<11)
-#define MPLP_SMART_OVERLAPS (1<<12)
+#define MPLP_PER_SAMPLE (1<<9)
+#define MPLP_SMART_OVERLAPS (1<<10)
 
+#define MPLP_PRINT_MAPQ_CHAR (1<<11)
+#define MPLP_PRINT_QPOS  (1<<12)
 #define MPLP_PRINT_QNAME (1<<13)
 #define MPLP_PRINT_FLAG  (1<<14)
 #define MPLP_PRINT_RNAME (1<<15)
@@ -143,6 +192,10 @@ static inline int pileup_seq(FILE *fp, const bam_pileup1_t *p, hts_pos_t pos,
 #define MPLP_PRINT_TLEN  (1<<21)
 #define MPLP_PRINT_SEQ   (1<<22)
 #define MPLP_PRINT_QUAL  (1<<23)
+#define MPLP_PRINT_MODS  (1<<24)
+#define MPLP_PRINT_QPOS5 (1<<25)
+
+#define MPLP_PRINT_LAST  (1<<26) // terminator for loop
 
 #define MPLP_MAX_DEPTH 8000
 #define MPLP_MAX_INDEL_DEPTH 250
@@ -157,7 +210,7 @@ typedef struct {
     void *bed, *rghash, *auxlist;
     int argc;
     char **argv;
-    char sep, empty;
+    char sep, empty, no_ins, no_ins_mods, no_del, no_ends;
     sam_global_args ga;
 } mplp_conf_t;
 
@@ -288,6 +341,23 @@ static int mplp_get_ref(mplp_aux_t *ma, int tid, char **ref, hts_pos_t *ref_len)
     return 1;
 }
 
+// Initialise and destroy the base modifier state data. This is called
+// as each new read is added or removed from the pileups.
+static
+int pileup_cd_create(void *data, const bam1_t *b, bam_pileup_cd *cd) {
+    int ret;
+    hts_base_mod_state *m = hts_base_mod_state_alloc();
+    ret = bam_parse_basemod(b, m);
+    cd->p = m;
+    return ret;
+}
+
+static
+int pileup_cd_destroy(void *data, const bam1_t *b, bam_pileup_cd *cd) {
+    hts_base_mod_state_free(cd->p);
+    return 0;
+}
+
 static void
 print_empty_pileup(FILE *fp, const mplp_conf_t *conf, const char *tname,
                    hts_pos_t pos, int n, const char *ref, hts_pos_t ref_len)
@@ -296,11 +366,9 @@ print_empty_pileup(FILE *fp, const mplp_conf_t *conf, const char *tname,
     fprintf(fp, "%s\t%"PRIhts_pos"\t%c", tname, pos+1, (ref && pos < ref_len)? ref[pos] : 'N');
     for (i = 0; i < n; ++i) {
         fputs("\t0\t*\t*", fp);
-        if (conf->flag & MPLP_PRINT_QPOS)
-            fputs("\t*", fp);
-        int flag_value = MPLP_PRINT_QNAME;
-        while(flag_value < MPLP_PRINT_QUAL + 1) {
-            if (conf->flag & flag_value)
+        int flag_value = MPLP_PRINT_MAPQ_CHAR;
+        while(flag_value < MPLP_PRINT_LAST) {
+            if (flag_value != MPLP_PRINT_MODS && (conf->flag & flag_value))
                 fputs("\t*", fp);
             flag_value <<= 1;
         }
@@ -389,7 +457,7 @@ static void group_smpl(mplp_pileup_t *m, bam_sample_t *sm, kstring_t *buf,
             if (id < 0 || id >= m->n) {
                 assert(q); // otherwise a bug
                 fprintf(samtools_stderr, "[%s] Read group %s used in file %s but absent from the header or an alignment missing read group.\n", __func__, (char*)q+1, fn[i]);
-                exit(EXIT_FAILURE);
+                samtools_exit(EXIT_FAILURE);
             }
             if (m->n_plp[id] == m->m_plp[id]) {
                 m->m_plp[id] = m->m_plp[id]? m->m_plp[id]<<1 : 8;
@@ -442,10 +510,11 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
 
     if (n == 0) {
         fprintf(samtools_stderr,"[%s] no input file/data given\n", __func__);
-        exit(EXIT_FAILURE);
+        samtools_exit(EXIT_FAILURE);
     }
 
     // read the header of each file in the list and initialize data
+    refs_t *refs = NULL;
     for (i = 0; i < n; ++i) {
         sam_hdr_t *h_tmp;
         data[i] = calloc(1, sizeof(mplp_aux_t));
@@ -453,23 +522,34 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
         if ( !data[i]->fp )
         {
             fprintf(samtools_stderr, "[%s] failed to open %s: %s\n", __func__, fn[i], strerror(errno));
-            exit(EXIT_FAILURE);
+            samtools_exit(EXIT_FAILURE);
         }
         if (hts_set_opt(data[i]->fp, CRAM_OPT_DECODE_MD, 0)) {
             fprintf(samtools_stderr, "Failed to set CRAM_OPT_DECODE_MD value\n");
-            exit(EXIT_FAILURE);
+            samtools_exit(EXIT_FAILURE);
         }
-        if (conf->fai_fname && hts_set_fai_filename(data[i]->fp, conf->fai_fname) != 0) {
-            fprintf(samtools_stderr, "[%s] failed to process %s: %s\n",
-                    __func__, conf->fai_fname, strerror(errno));
-            exit(EXIT_FAILURE);
+
+        if (!refs && conf->fai_fname) {
+            if (hts_set_fai_filename(data[i]->fp, conf->fai_fname) != 0) {
+                fprintf(samtools_stderr, "[%s] failed to process %s: %s\n",
+                        __func__, conf->fai_fname, strerror(errno));
+                samtools_exit(EXIT_FAILURE);
+            }
+            refs = cram_get_refs(data[i]->fp);
+        } else if (conf->fai_fname) {
+            if (hts_set_opt(data[i]->fp, CRAM_OPT_SHARED_REF, refs) != 0) {
+                fprintf(samtools_stderr, "[%s] failed to process %s: %s\n",
+                        __func__, conf->fai_fname, strerror(errno));
+                samtools_exit(EXIT_FAILURE);
+            }
         }
+
         data[i]->conf = conf;
         data[i]->ref = &mp_ref;
         h_tmp = sam_hdr_read(data[i]->fp);
         if ( !h_tmp ) {
             fprintf(samtools_stderr,"[%s] fail to read the header of %s\n", __func__, fn[i]);
-            exit(EXIT_FAILURE);
+            samtools_exit(EXIT_FAILURE);
         }
         bam_smpl_add(sm, fn[i], (conf->flag&MPLP_IGNORE_RG)? 0 : sam_hdr_str(h_tmp));
         if (conf->flag & MPLP_BCF) {
@@ -487,11 +567,11 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
 
             if (idx == NULL) {
                 fprintf(samtools_stderr, "[%s] fail to load index for %s\n", __func__, fn[i]);
-                exit(EXIT_FAILURE);
+                samtools_exit(EXIT_FAILURE);
             }
             if ( (data[i]->iter=sam_itr_querys(idx, h_tmp, conf->reg)) == 0) {
                 fprintf(samtools_stderr, "[E::%s] fail to parse region '%s' with %s\n", __func__, conf->reg, fn[i]);
-                exit(EXIT_FAILURE);
+                samtools_exit(EXIT_FAILURE);
             }
             if (i == 0) beg0 = data[i]->iter->beg, end0 = data[i]->iter->end, tid0 = data[i]->iter->tid;
             hts_idx_destroy(idx);
@@ -529,8 +609,9 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
         bcf_fp = bcf_open(conf->output_fname? conf->output_fname : "-", mode);
         if (bcf_fp == NULL) {
             fprintf(samtools_stderr, "[%s] failed to write to %s: %s\n", __func__, conf->output_fname? conf->output_fname : "standard output", strerror(errno));
-            exit(EXIT_FAILURE);
+            samtools_exit(EXIT_FAILURE);
         }
+        autoflush_if_stdout(bcf_fp, conf->output_fname);
 
         // BCF header creation
         bcf_hdr = bcf_hdr_init("w");
@@ -613,7 +694,7 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
         if (bcf_hdr_write(bcf_fp, bcf_hdr) != 0) {
             print_error_errno("mpileup", "Failed to write VCF/BCF header to \"%s\"",
                               conf->output_fname? conf->output_fname : "standard output");
-            exit(EXIT_FAILURE);
+            samtools_exit(EXIT_FAILURE);
         }
         // End of BCF header creation
 
@@ -652,12 +733,16 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
 
         if (pileup_fp == NULL) {
             fprintf(samtools_stderr, "[%s] failed to write to %s: %s\n", __func__, conf->output_fname, strerror(errno));
-            exit(EXIT_FAILURE);
+            samtools_exit(EXIT_FAILURE);
         }
     }
 
     // init pileup
     iter = bam_mplp_init(n, mplp_func, (void**)data);
+    if (conf->flag & MPLP_PRINT_MODS) {
+        bam_mplp_constructor(iter, pileup_cd_create);
+        bam_mplp_destructor(iter, pileup_cd_destroy);
+    }
     if ( conf->flag & MPLP_SMART_OVERLAPS ) bam_mplp_init_overlaps(iter);
     if ( !conf->max_depth ) {
         max_depth = INT_MAX;
@@ -698,7 +783,7 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
             if (bcf_write1(bcf_fp, bcf_hdr, bcf_rec) != 0) {
                 print_error_errno("mpileup", "Failed to write VCF/BCF record to \"%s\"",
                                   conf->output_fname?conf->output_fname:"standard output");
-                exit(EXIT_FAILURE);
+                samtools_exit(EXIT_FAILURE);
             }
             // call indels; todo: subsampling with total_depth>max_indel_depth instead of ignoring?
             if (!(conf->flag&MPLP_NO_INDEL) && total_depth < max_indel_depth && bcf_call_gap_prep(gplp.n, gplp.n_plp, gplp.plp, pos, bca, ref, rghash) >= 0)
@@ -712,7 +797,7 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
                     if (bcf_write1(bcf_fp, bcf_hdr, bcf_rec) != 0) {
                         print_error_errno("mpileup", "Failed to write VCF/BCF record to \"%s\"",
                                           conf->output_fname?conf->output_fname:"standard output");
-                        exit(EXIT_FAILURE);
+                        samtools_exit(EXIT_FAILURE);
                     }
                 }
             }
@@ -759,11 +844,10 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
                 fprintf(pileup_fp, "\t%d\t", cnt);
                 if (n_plp[i] == 0) {
                     fputs("*\t*", pileup_fp);
-                    if (conf->flag & MPLP_PRINT_QPOS)
-                        fputs("\t*", pileup_fp);
-                    int flag_value = MPLP_PRINT_QNAME;
-                    while(flag_value < MPLP_PRINT_QUAL + 1) {
-                        if (conf->flag & flag_value)
+                    int flag_value = MPLP_PRINT_MAPQ_CHAR;
+                    while(flag_value < MPLP_PRINT_LAST) {
+                        if (flag_value != MPLP_PRINT_MODS
+                            && (conf->flag & flag_value))
                             fputs("\t*", pileup_fp);
                         flag_value <<= 1;
                     }
@@ -782,7 +866,10 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
                             : 0;
                         if (c >= conf->min_baseQ) {
                             n++;
-                            if (pileup_seq(pileup_fp, plp[i] + j, pos, ref_len, ref, &ks, conf->rev_del) < 0) {
+                            if (pileup_seq(pileup_fp, plp[i] + j, pos, ref_len,
+                                           ref, &ks, conf->rev_del,
+                                           conf->no_ins, conf->no_ins_mods,
+                                           conf->no_del, conf->no_ends) < 0) {
                                 ret = 1;
                                 goto fail;
                             }
@@ -807,27 +894,11 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
                     }
                     if (!n) putc('*', pileup_fp);
 
-                    /* Print mpileup positions */
-                    if (conf->flag & MPLP_PRINT_QPOS) {
-                        n = 0;
-                        putc('\t', pileup_fp);
-                        for (j = 0; j < n_plp[i]; ++j) {
-                            const bam_pileup1_t *p = plp[i] + j;
-                            int c = p->qpos < p->b->core.l_qseq
-                                    ? bam_get_qual(p->b)[p->qpos]
-                                                         : 0;
-                            if ( c < conf->min_baseQ ) continue;
-                            if (n > 0) putc(',', pileup_fp);
-                            n++;
-                            fprintf(pileup_fp, "%d", p->qpos + 1);
-                        }
-                        if (!n) putc('*', pileup_fp);
-                    }
-
                     /* Print selected columns */
-                    int flag_value = MPLP_PRINT_QNAME;
-                    while(flag_value < MPLP_PRINT_QUAL + 1) {
-                        if (conf->flag & flag_value) {
+                    int flag_value = MPLP_PRINT_MAPQ_CHAR;
+                    while(flag_value < MPLP_PRINT_LAST) {
+                        if (flag_value != MPLP_PRINT_MODS
+                            && (conf->flag & flag_value)) {
                             n = 0;
                             putc('\t', pileup_fp);
                             for (j = 0; j < n_plp[i]; ++j) {
@@ -836,10 +907,27 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
                                     ? bam_get_qual(p->b)[p->qpos]
                                     : 0;
                                 if ( c < conf->min_baseQ ) continue;
-                                if (n > 0 && flag_value != MPLP_PRINT_MAPQ) putc(',', pileup_fp);
+                                if (n > 0 && flag_value != MPLP_PRINT_MAPQ_CHAR) putc(',', pileup_fp);
                                 n++;
 
                                 switch (flag_value) {
+                                case MPLP_PRINT_MAPQ_CHAR:
+                                    c = p->b->core.qual + 33;
+                                    if (c > 126) c = 126;
+                                    putc(c, pileup_fp);
+                                    break;
+                                case MPLP_PRINT_QPOS:
+                                    // query position in current orientation
+                                    fprintf(pileup_fp, "%d", p->qpos + 1);
+                                    break;
+                                case MPLP_PRINT_QPOS5: {
+                                    // query position in 5' to 3' orientation
+                                    int pos5 = bam_is_rev(p->b)
+                                        ? p->b->core.l_qseq-p->qpos + p->is_del
+                                        : p->qpos + 1;
+                                    fprintf(pileup_fp, "%d", pos5);
+                                    break;
+                                }
                                 case MPLP_PRINT_QNAME:
                                     fputs(bam_get_qname(p->b), pileup_fp);
                                     break;
@@ -856,9 +944,7 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
                                     fprintf(pileup_fp, "%"PRId64, (int64_t) p->b->core.pos + 1);
                                     break;
                                 case MPLP_PRINT_MAPQ:
-                                    c = p->b->core.qual + 33;
-                                    if (c > 126) c = 126;
-                                    putc(c, pileup_fp);
+                                    fprintf(pileup_fp, "%d", p->b->core.qual);
                                     break;
                                 case MPLP_PRINT_RNEXT:
                                     if (p->b->core.mtid >= 0)
@@ -932,6 +1018,12 @@ static int mpileup(mplp_conf_t *conf, int n, char **fn, char **fn_idx)
         }
     }
 
+    if (ret < 0) {
+        print_error("mpileup", "error reading from input file");
+        ret = EXIT_FAILURE;
+        goto fail;
+    }
+
     if (conf->all && !(conf->flag & MPLP_BCF)) {
         // Handle terminating region
         if (last_tid < 0 && conf->reg && conf->all > 1) {
@@ -959,6 +1051,7 @@ fail:
     bcf_destroy1(bcf_rec);
     if (bcf_fp)
     {
+        release_autoflush(bcf_fp);
         hts_close(bcf_fp);
         bcf_hdr_destroy(bcf_hdr);
         bcf_call_destroy(bca);
@@ -1073,7 +1166,7 @@ int parse_format_flag(const char *str)
         else
         {
             fprintf(samtools_stderr,"Could not parse tag \"%s\" in \"%s\"\n", tags[i], str);
-            exit(EXIT_FAILURE);
+            samtools_exit(EXIT_FAILURE);
         }
         free(tags[i]);
     }
@@ -1112,25 +1205,33 @@ static void print_usage(FILE *fp, const mplp_conf_t *mplp)
     fprintf(fp,
 "  -r, --region REG        region in which pileup is generated\n"
 "  -R, --ignore-RG         ignore RG tags (one BAM = one sample)\n"
-"  --rf, --incl-flags STR|INT  required flags: skip reads with mask bits unset [%s]\n", tmp_require);
+"  --rf, --incl-flags STR|INT  required flags: include reads with any of the mask bits set [%s]\n", tmp_require);
     fprintf(fp,
-"  --ff, --excl-flags STR|INT  filter flags: skip reads with mask bits set\n"
+"  --ff, --excl-flags STR|INT  filter flags: skip reads with any of the mask bits set\n"
 "                                            [%s]\n", tmp_filter);
     fprintf(fp,
 "  -x, --ignore-overlaps   disable read-pair overlap detection\n"
 "  -X, --customized-index  use customized index files\n" // -X flag for index filename
 "\n"
 "Output options:\n"
-"  -o, --output FILE       write output to FILE [standard output]\n"
-"  -O, --output-BP         output base positions on reads\n"
-"  -s, --output-MQ         output mapping quality\n"
-"      --output-QNAME      output read names\n"
-"      --output-extra STR  output extra read fields and read tag values\n"
-"      --output-sep CHAR   set the separator character for tag lists [,]\n"
-"      --output-empty CHAR set the no value character for tag lists [*]\n"
-"      --reverse-del       use '#' character for deletions on the reverse strand\n"
-"  -a                      output all positions (including zero depth)\n"
-"  -a -a (or -aa)          output absolutely all positions, including unused ref. sequences\n"
+"  -o, --output FILE        write output to FILE [standard output]\n"
+"  -O, --output-BP          output base positions on reads, current orientation\n"
+"      --output-BP-5        output base positions on reads, 5' to 3' orientation\n"
+"  -M, --output-mods        output base modifications\n"
+"  -s, --output-MQ          output mapping quality\n"
+"      --output-QNAME       output read names\n"
+"      --output-extra STR   output extra read fields and read tag values\n"
+"      --output-sep CHAR    set the separator character for tag lists [,]\n"
+"      --output-empty CHAR  set the no value character for tag lists [*]\n"
+"      --no-output-ins      skip insertion sequence after +NUM\n"
+"                           Use twice for complete insertion removal\n"
+"      --no-output-ins-mods don't display base modifications within insertions\n"
+"      --no-output-del      skip deletion sequence after -NUM\n"
+"                           Use twice for complete deletion removal\n"
+"      --no-output-ends     remove ^MQUAL and $ markup in sequence column\n"
+"      --reverse-del        use '#' character for deletions on the reverse strand\n"
+"  -a                       output all positions (including zero depth)\n"
+"  -a -a (or -aa)           output absolutely all positions, including unused ref. sequences\n"
 "\n"
 "Generic options:\n");
     sam_global_opt_help(fp, "-.--.--.");
@@ -1209,8 +1310,11 @@ int bam_mpileup(int argc, char *argv[])
         {"bcf", no_argument, NULL, 'g'},
         {"VCF", no_argument, NULL, 'v'},
         {"vcf", no_argument, NULL, 'v'},
+        {"output-mods", no_argument, NULL, 'M'},
         {"output-BP", no_argument, NULL, 'O'},
         {"output-bp", no_argument, NULL, 'O'},
+        {"output-BP-5", no_argument, NULL, 14},
+        {"output-bp-5", no_argument, NULL, 14},
         {"output-MQ", no_argument, NULL, 's'},
         {"output-mq", no_argument, NULL, 's'},
         {"output-tags", required_argument, NULL, 't'},
@@ -1229,10 +1333,14 @@ int bam_mpileup(int argc, char *argv[])
         {"output-extra", required_argument, NULL, 7},
         {"output-sep", required_argument, NULL, 8},
         {"output-empty", required_argument, NULL, 9},
+        {"no-output-ins", no_argument, NULL, 10},
+        {"no-output-ins-mods", no_argument, NULL, 11},
+        {"no-output-del", no_argument, NULL, 12},
+        {"no-output-ends", no_argument, NULL, 13},
         {NULL, 0, NULL, 0}
     };
 
-    while ((c = getopt_long(argc, argv, "Agf:r:l:q:Q:uRC:BDSd:L:b:P:po:e:h:Im:F:EG:6OsVvxXt:a",lopts,NULL)) >= 0) {
+    while ((c = getopt_long(argc, argv, "Agf:r:l:q:Q:uRC:BDSd:L:b:P:po:e:h:Im:F:EG:6OsVvxXt:aM",lopts,NULL)) >= 0) {
         switch (c) {
         case 'x': mplp.flag &= ~MPLP_SMART_OVERLAPS; break;
         case  1 :
@@ -1255,6 +1363,10 @@ int bam_mpileup(int argc, char *argv[])
             break;
         case 8: mplp.sep = optarg[0]; break;
         case 9: mplp.empty = optarg[0]; break;
+        case 10: mplp.no_ins++; break;
+        case 11: mplp.no_ins_mods = 1; break;
+        case 12: mplp.no_del++; break;
+        case 13: mplp.no_ends = 1; break;
         case 'f':
             mplp.fai = fai_load(optarg);
             if (mplp.fai == NULL) return 1;
@@ -1283,8 +1395,16 @@ int bam_mpileup(int argc, char *argv[])
         case 'E': mplp.flag |= MPLP_REDO_BAQ; break;
         case '6': mplp.flag |= MPLP_ILLUMINA13; break;
         case 'R': mplp.flag |= MPLP_IGNORE_RG; break;
-        case 's': mplp.flag |= MPLP_PRINT_MAPQ; break;
-        case 'O': mplp.flag |= MPLP_PRINT_QPOS; break;
+        case 's': mplp.flag |= MPLP_PRINT_MAPQ_CHAR; break;
+        case 'O':
+            if (!(mplp.flag & MPLP_PRINT_QPOS5))
+                mplp.flag |= MPLP_PRINT_QPOS;
+            break;
+        case  14:
+            mplp.flag |=  MPLP_PRINT_QPOS5;
+            mplp.flag &= ~MPLP_PRINT_QPOS;
+            break;
+        case 'M': mplp.flag |= MPLP_PRINT_MODS; break;
         case 'C': mplp.capQ_thres = atoi(optarg); break;
         case 'q': mplp.min_mq = atoi(optarg); break;
         case 'Q': mplp.min_baseQ = atoi(optarg); break;
