@@ -382,7 +382,7 @@ int cram_reheader_inplace3(cram_fd *fd, sam_hdr_t *h, const char *arg_list,
     cram_block_append(b, (void *)sam_hdr_str(cram_h), header_len);
     cram_block_update_size(b);
 
-    cram_compress_block(fd, b, NULL, -1, -1);
+    cram_compress_block(fd, b, NULL, -1, 9);
 
     if (hseek(cram_fd_get_fp(fd), 26, SEEK_SET) != 26)
         goto err;
@@ -444,7 +444,7 @@ static void usage(FILE *fp, int ret) {
            "    -i, --in-place      Modify the CRAM file directly, if possible.\n"
            "                        (Defaults to outputting to samtools_stdout.)\n"
            "    -c, --command CMD   Pass the header in SAM format to external program CMD.\n");
-    exit(ret);
+    samtools_exit(ret);
 }
 
 static sam_hdr_t* external_reheader(samFile* in, const char* external) {
@@ -533,7 +533,7 @@ cleanup:
     return h;
 }
 
-int main_reheader(int argc, char *argv[])
+int samtools_main_reheader(int argc, char *argv[])
 {
     int inplace = 0, r, no_pg = 0, c, skip_header = 0;
     sam_hdr_t *h;
